use std::{
	sync::mpsc::{self, Receiver, Sender},
	thread,
};

use fltk::{
	app,
	enums::{Color, FrameType},
	misc::Chart,
	prelude::{GroupExt, WidgetExt, WindowExt},
	window::Window,
};
use fltk_theme::{color_themes, ColorTheme};
use rand::Rng;

mod bubble;
mod merge;
mod quick;

const SIZE: u32 = 100;
const MAX_VALUE: f64 = 100.0;

type Sorter = fn(&mut SortVisualizer);

enum SortMessage {
	Data(Vec<f64>),
	Done,
}

#[derive(Debug)]
struct SortVisualizer<'a> {
	arr: Vec<f64>,
	name: &'a str,
	chart: Chart,
	sorter: Sorter,
	tx: Sender<SortMessage>,
	rx: Receiver<SortMessage>,
}

impl<'a> SortVisualizer<'a> {
	fn new(id: usize, name: &'a str, arr: Vec<f64>, sorter: Sorter) -> Self {
		let (tx, rx) = mpsc::channel();
		let y = 300 * (id) as i32;

		let mut chart = Chart::default();
		chart.set_pos(0, y);
		chart.set_size(500, 250);
		chart.set_frame(FrameType::FlatBox);
		chart.set_bounds(0.0, MAX_VALUE);
		chart.set_label(name);

		Self {
			arr,
			name,
			chart,
			sorter,
			tx,
			rx,
		}
	}

	fn sort(&mut self) {
		(self.sorter)(self);
	}

	fn draw(&mut self, arr: Vec<f64>) {
		self.chart.clear();
		for i in &arr {
			self.chart.add(*i, "", Color::Cyan);
		}
	}

	fn complete(&mut self, rank: u8) {
		self.chart
			.set_label(&format!("{} - Complete - Rank #{}", self.name, rank));
	}
}

fn main() {
	let arr: Vec<f64> = (1..SIZE)
		.map(|_| rand::rng().random_range(1.0..MAX_VALUE))
		.collect();

	let sorts: [(&str, Sorter); 3] = [
		("Quick Sort", quick::sort),
		("Bubble Sort", bubble::sort),
		("Merge Sort", merge::sort),
	];

	let app = app::App::default().with_scheme(app::Scheme::Gtk);
	let theme = ColorTheme::new(&color_themes::fleet::DARK1);
	theme.apply();

	let mut window = Window::default()
		.with_size(640, 480)
		.with_label("FLTK Visualizer");
	window.size_range(640, 480, 0, 0);

	let mut visualizers = vec![];

	for (i, sort) in sorts.into_iter().enumerate() {
		visualizers.push(SortVisualizer::new(i, sort.0, arr.clone(), sort.1));
	}

	thread::scope(|scope| {
		for sorter in &mut visualizers {
			scope.spawn(move || {
				sorter.sort();
			});
		}
	});

	window.end();
	window.show();

	let mut rank = 0;

	app::add_idle3(move |_| {
		for sorter in &mut visualizers {
			let data = sorter.rx.try_recv();
			if let Ok(data) = data {
				match data {
					SortMessage::Data(arr) => sorter.draw(arr),
					SortMessage::Done => {
						rank += 1;
						sorter.complete(rank);
					}
				}
			}
		}

		window.redraw();
		app::sleep(0.016);
	});

	app.run().unwrap();
}
